

from mocat.src.scenarios.twodim.vectorise import TwoDimToyScenario

from mocat.src.scenarios.twodim.toy_examples import Gaussian
from mocat.src.scenarios.twodim.toy_examples import Banana
from mocat.src.scenarios.twodim.toy_examples import NealFunnel
from mocat.src.scenarios.twodim.toy_examples import GaussianMixture
from mocat.src.scenarios.twodim.toy_examples import DoubleWell
from mocat.src.scenarios.twodim.toy_examples import Rastrigin

