
from mocat import twodim

from mocat.src.core import Scenario

from mocat.src.scenarios.toy_examples import Gaussian
from mocat.src.scenarios.toy_examples import GaussianMixture
from mocat.src.scenarios.toy_examples import DoubleWell
from mocat.src.scenarios.toy_examples import Rastrigin



